"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var StringUtils_1 = require("./StringUtils");
var NodeUtils = /** @class */ (function () {
    function NodeUtils() {
    }
    NodeUtils.nextValidTextNode = function (node, including) {
        node = including ? node : this.nextNode(node);
        while (node) {
            var valid = this.firstValidTextNode(node);
            if (valid) {
                return valid;
            }
            node = this.nextNode(node);
        }
        throw new Error('No Valid Text Node');
    };
    NodeUtils.prevValidTextNode = function (node, including) {
        node = including ? node : this.prevNode(node);
        while (node) {
            var valid = this.lastValidTextNode(node);
            if (valid) {
                return valid;
            }
            node = this.prevNode(node);
        }
        throw new Error('No Valid Text Node');
    };
    NodeUtils.nextNode = function (node) {
        var next = node.nextSibling;
        while (!next) {
            node = node.parentNode;
            next = node.nextSibling;
        }
        return next;
    };
    NodeUtils.prevNode = function (node) {
        var prev = node.previousSibling;
        while (!prev) {
            node = node.parentNode;
            prev = node.previousSibling;
        }
        return prev;
    };
    NodeUtils.firstValidTextNode = function (node) {
        if (this.isValidTextNode(node)) {
            return node;
        }
        else if (node.nodeType === 1 && !this.isSkippable(node)) {
            var ordered = Array.from(node.childNodes);
            for (var _i = 0, ordered_1 = ordered; _i < ordered_1.length; _i++) {
                var childNode = ordered_1[_i];
                var valid = this.firstValidTextNode(childNode);
                if (valid) {
                    return valid;
                }
            }
        }
        return null;
    };
    NodeUtils.lastValidTextNode = function (node) {
        if (this.isValidTextNode(node)) {
            return node;
        }
        else if (node.nodeType === 1 && !this.isSkippable(node)) {
            var reversed = Array.from(node.childNodes).reverse();
            for (var _i = 0, reversed_1 = reversed; _i < reversed_1.length; _i++) {
                var childNode = reversed_1[_i];
                var valid = this.lastValidTextNode(childNode);
                if (valid) {
                    return valid;
                }
            }
        }
        return null;
    };
    NodeUtils.isValidTextNode = function (node) {
        return node.nodeType === 3 && node.length > 0 && /\S/.test(node.data);
    };
    NodeUtils.getTextNodesBetween = function (startNode, endNode) {
        var result = [];
        var node = startNode;
        result.push(node);
        while (node !== endNode) {
            node = this.nextValidTextNode(node);
            result.push(node);
        }
        return result;
    };
    NodeUtils.getValidTextNodes = function (node) {
        var _this = this;
        if (this.isValidTextNode(node)) {
            return [node];
        }
        else if (node.nodeType === 1 && !this.isSkippable(node)) {
            return Array.from(node.childNodes).reduce(function (nodes, child) {
                return nodes.concat(_this.getValidTextNodes(child));
            }, []);
        }
        return [];
    };
    NodeUtils.getContentTo = function (end, endOffset, xdoc) {
        var first = this.firstValidTextNode(xdoc.root);
        var nodes = this.getTextNodesBetween(first, end);
        var tail = end.substringData(0, endOffset);
        var content = nodes.slice(0, nodes.length - 1).map(function (n) { return n.data.trim(); }).join('') + tail;
        return StringUtils_1.StringUtils.compact(content);
    };
    NodeUtils.isSkippable = function (node) {
        return this.elementsToSkip.test(node, null);
    };
    NodeUtils.elementsToSkip = {
        elements: [
            'applet',
            'area',
            'base',
            'basefont',
            'bdo',
            'button',
            'frame',
            'frameset',
            'iframe',
            'head',
            'hr',
            'img',
            'input',
            'link',
            'map',
            'meta',
            'noframes',
            'noscript',
            'optgroup',
            'option',
            'param',
            'script',
            'select',
            'style',
            'textarea',
            'title',
        ],
        /**
         * @param {Element} node
         * @param {string[]|null} list
         * @return {boolean}
         */
        test: function (node, list) {
            var elements = list || this.elements;
            return elements.indexOf(node.tagName.toLowerCase()) > -1;
        },
    };
    return NodeUtils;
}());
exports.NodeUtils = NodeUtils;
//# sourceMappingURL=NodeUtils.js.map