"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var NodeUtils_1 = require("./NodeUtils");
var StringUtils_1 = require("./StringUtils");
var RangeUtils = /** @class */ (function () {
    function RangeUtils() {
    }
    RangeUtils.trim = function (range) {
        var oldStart = range.startContainer;
        var newStart = NodeUtils_1.NodeUtils.nextValidTextNode(oldStart, true);
        var newStartOffset = StringUtils_1.StringUtils.indexOfNthNotEmptyChar(newStart.data, 0);
        if (oldStart === newStart) {
            newStartOffset = Math.max(range.startOffset, newStartOffset);
            if (newStartOffset > StringUtils_1.StringUtils.indexOfNthNotEmptyCharFromEnd(newStart.data, 0)) {
                // start at the tail of a valid text node
                newStart = NodeUtils_1.NodeUtils.nextValidTextNode(newStart, false);
                newStartOffset = StringUtils_1.StringUtils.indexOfNthNotEmptyChar(newStart.data, 0);
            }
            else {
                // start at midden of a valid text node
                var head = newStart.substringData(newStartOffset, newStart.length - newStartOffset);
                // num of whitespaces at head
                var headSpaces = StringUtils_1.StringUtils.indexOfNthNotEmptyChar(head, 0);
                newStartOffset += headSpaces; // skip whitespaces
            }
        }
        var oldEnd = range.endContainer;
        var newEnd = NodeUtils_1.NodeUtils.prevValidTextNode(oldEnd, true);
        var newEndOffset = StringUtils_1.StringUtils.indexOfNthNotEmptyCharFromEnd(newEnd.data, 0) + 1;
        if (oldEnd === newEnd) {
            newEndOffset = Math.min(range.endOffset, newEndOffset);
            if (newEndOffset <= StringUtils_1.StringUtils.indexOfNthNotEmptyChar(newEnd.data, 0)) {
                // end at the head of a valid text node
                newEnd = NodeUtils_1.NodeUtils.prevValidTextNode(newEnd, false);
                newEndOffset = StringUtils_1.StringUtils.indexOfNthNotEmptyCharFromEnd(newEnd.data, 0) + 1;
            }
            else {
                // end at midden of a valid text node
                var tail = newEnd.substringData(0, newEndOffset);
                // num of whitespaces at tail
                var tailSpaces = tail.length - 1 - StringUtils_1.StringUtils.indexOfNthNotEmptyCharFromEnd(tail, 0);
                newEndOffset -= tailSpaces; // skip whitespaces
            }
        }
        range.setStart(newStart, newStartOffset); // start with a non-whitespace char
        range.setEnd(newEnd, newEndOffset); // end with a non-whitespace char
        return range;
    };
    RangeUtils.extractTextNodes = function (range) {
        var sContainer = range.startContainer;
        var eContainer = range.endContainer;
        var sOffset = range.startOffset;
        var eOffset = range.endOffset;
        if (eContainer.data.length !== eOffset) {
            eContainer.splitText(eOffset);
        }
        if (sOffset !== 0) {
            var newStartContainer = sContainer.splitText(sOffset);
            if (eContainer === sContainer) {
                eContainer = newStartContainer;
                eOffset = eOffset - sOffset;
            }
            sContainer = newStartContainer;
            sOffset = 0;
        }
        range.setStart(sContainer, sOffset);
        range.setEnd(eContainer, eOffset);
        return NodeUtils_1.NodeUtils.getTextNodesBetween(sContainer, eContainer);
    };
    return RangeUtils;
}());
exports.default = RangeUtils;
//# sourceMappingURL=RangeUtils.js.map