"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var XSelection_1 = require("./XSelection");
var StringUtils_1 = require("./StringUtils");
var NodeUtils_1 = require("./NodeUtils");
var XDocument = /** @class */ (function () {
    function XDocument(root) {
        this.selection = null;
        this.root = root;
        this.doc = (root instanceof Document ? root : root.ownerDocument);
        this.win = this.doc.defaultView;
        this.refresh();
    }
    XDocument.from = function (root) {
        if (root === void 0) { root = document.body; }
        return new XDocument(root);
    };
    XDocument.prototype.refresh = function () {
        var nodes = NodeUtils_1.NodeUtils.getValidTextNodes(this.root);
        var position = 0;
        var text = '';
        nodes.forEach(function (node) {
            node.startPosition = position;
            var nodeText = StringUtils_1.StringUtils.compact(node.data);
            text += nodeText;
            position += nodeText.length;
            node.endPosition = position - 1;
        });
        this.text = text;
        this.nodes = nodes;
    };
    /**
     * @see {@code XSelection.fromText}
     */
    XDocument.prototype.select = function (optText, optNth, optSelect) {
        if (optNth === void 0) { optNth = 1; }
        if (optSelect === void 0) { optSelect = false; }
        if (optText && optText.trim()) {
            if (this.selection && !this.selection.isEmpty()) {
                this.selection.empty();
            }
            this.selection = this.fromText(optText.trim(), optNth, optSelect);
        }
        else {
            var selection = this.win.getSelection();
            if (selection) {
                this.selection = this.fromSelection(selection);
            }
        }
        return this.selection;
    };
    /**
     * @see {@code XSelection.fromSelection}
     */
    XDocument.prototype.fromSelection = function (selection) {
        return this.selection = XSelection_1.XSelection.fromSelection(selection, this);
    };
    /**
     * @see {@code XSelection.fromText}
     */
    XDocument.prototype.fromText = function (text, nth, optSelect) {
        if (nth === void 0) { nth = 1; }
        if (optSelect === void 0) { optSelect = false; }
        return this.selection = this.fromTextIndex({ text: text, nth: nth }, optSelect);
    };
    /**
     * @see {@code XSelection.fromText}
     */
    XDocument.prototype.fromTextIndex = function (index, optSelect) {
        if (optSelect === void 0) { optSelect = false; }
        return this.selection = XSelection_1.XSelection.fromText(index, optSelect, this);
    };
    /**
     * @see {@code XSelection.fromTextRange}
     */
    XDocument.prototype.fromTextRange = function (range, optSelect) {
        if (optSelect === void 0) { optSelect = false; }
        return this.selection = XSelection_1.XSelection.fromTextRange(range, optSelect, this);
    };
    XDocument.prototype.normalize = function () {
        var _this = this;
        this.root.normalize();
        var end = 0;
        this.nodes.reverse().forEach(function (node, index) {
            if (!node.parentNode) {
                end = node.endPosition;
                _this.nodes.splice(_this.nodes.length - index, 1);
            }
            else if (end > 0) {
                node.endPosition = end;
            }
        });
    };
    return XDocument;
}());
exports.XDocument = XDocument;
//# sourceMappingURL=XDocument.js.map