var StringUtils = /** @class */ (function () {
    function StringUtils() {
    }
    /**
     * the position of the nth occurrence of searchString
     * @param {string} source
     * @param {string} searchString
     * @param {number} nth = 1
     * @return {number} start position of the nth occurrence of <em>searchString</em>
     */
    StringUtils.indexOf = function (source, searchString, nth) {
        if (nth === void 0) { nth = 1; }
        var times = 0;
        var index = -2;
        while (times < nth && index !== -1) {
            index = source.indexOf(searchString, index + 1);
            times++;
        }
        return index;
    };
    StringUtils.indexOfNthNotEmptyChar = function (source, nth) {
        var i = 0;
        for (var n = -1; n <= nth && i < source.length; i++) {
            if (!/\s+/m.test(source[i])) {
                n++;
                if (n === nth) {
                    return i;
                }
            }
        }
        return -1;
    };
    StringUtils.indexOfNthNotEmptyCharFromEnd = function (source, nth) {
        var i = source.length - 1;
        for (var n = -1; n <= nth && i > -1; i--) {
            if (!/\s+/m.test(source[i])) {
                n++;
                if (n === nth) {
                    return i;
                }
            }
        }
        return -1;
    };
    /**
     * the occurrences(include the positions) of searchString
     * @param {string} source
     * @param {string} searchString
     * @returns {Array<ITextIndex>}
     */
    StringUtils.find = function (source, searchString) {
        var times = 0;
        var position = -2;
        var result = [];
        if (searchString.length < 1) {
            return result;
        }
        var text = StringUtils.compact(searchString);
        while (position !== -1) {
            position = source.indexOf(searchString, position + 1);
            if (position < 0) {
                break;
            }
            result.push({ nth: ++times, text: text, position: position });
        }
        return result;
    };
    /**
     * @return {string} compacted string
     */
    StringUtils.compact = function (source) {
        return source.replace(/\s+/gm, '');
    };
    return StringUtils;
}());

var NodeUtils = /** @class */ (function () {
    function NodeUtils() {
    }
    NodeUtils.nextValidTextNode = function (node, including) {
        node = including ? node : this.nextNode(node);
        while (node) {
            var valid = this.firstValidTextNode(node);
            if (valid) {
                return valid;
            }
            node = this.nextNode(node);
        }
        throw new Error('No Valid Text Node');
    };
    NodeUtils.prevValidTextNode = function (node, including) {
        node = including ? node : this.prevNode(node);
        while (node) {
            var valid = this.lastValidTextNode(node);
            if (valid) {
                return valid;
            }
            node = this.prevNode(node);
        }
        throw new Error('No Valid Text Node');
    };
    NodeUtils.nextNode = function (node) {
        var next = node.nextSibling;
        while (!next) {
            node = node.parentNode;
            next = node.nextSibling;
        }
        return next;
    };
    NodeUtils.prevNode = function (node) {
        var prev = node.previousSibling;
        while (!prev) {
            node = node.parentNode;
            prev = node.previousSibling;
        }
        return prev;
    };
    NodeUtils.firstValidTextNode = function (node) {
        if (this.isValidTextNode(node)) {
            return node;
        }
        else if (node.nodeType === 1 && !this.isSkippable(node)) {
            var ordered = Array.from(node.childNodes);
            for (var _i = 0, ordered_1 = ordered; _i < ordered_1.length; _i++) {
                var childNode = ordered_1[_i];
                var valid = this.firstValidTextNode(childNode);
                if (valid) {
                    return valid;
                }
            }
        }
        return null;
    };
    NodeUtils.lastValidTextNode = function (node) {
        if (this.isValidTextNode(node)) {
            return node;
        }
        else if (node.nodeType === 1 && !this.isSkippable(node)) {
            var reversed = Array.from(node.childNodes).reverse();
            for (var _i = 0, reversed_1 = reversed; _i < reversed_1.length; _i++) {
                var childNode = reversed_1[_i];
                var valid = this.lastValidTextNode(childNode);
                if (valid) {
                    return valid;
                }
            }
        }
        return null;
    };
    NodeUtils.isValidTextNode = function (node) {
        return node.nodeType === 3 && node.length > 0 && /\S/.test(node.data);
    };
    NodeUtils.getTextNodesBetween = function (startNode, endNode) {
        var result = [];
        var node = startNode;
        result.push(node);
        while (node !== endNode) {
            node = this.nextValidTextNode(node);
            result.push(node);
        }
        return result;
    };
    NodeUtils.getValidTextNodes = function (node) {
        var _this = this;
        if (this.isValidTextNode(node)) {
            return [node];
        }
        else if (node.nodeType === 1 && !this.isSkippable(node)) {
            return Array.from(node.childNodes).reduce(function (nodes, child) {
                return nodes.concat(_this.getValidTextNodes(child));
            }, []);
        }
        return [];
    };
    NodeUtils.getContentTo = function (end, endOffset, xdoc) {
        var first = this.firstValidTextNode(xdoc.root);
        var nodes = this.getTextNodesBetween(first, end);
        var tail = end.substringData(0, endOffset);
        var content = nodes.slice(0, nodes.length - 1).map(function (n) { return n.data.trim(); }).join('') + tail;
        return StringUtils.compact(content);
    };
    NodeUtils.isSkippable = function (node) {
        return this.elementsToSkip.test(node, null);
    };
    NodeUtils.elementsToSkip = {
        elements: [
            'applet',
            'area',
            'base',
            'basefont',
            'bdo',
            'button',
            'frame',
            'frameset',
            'iframe',
            'head',
            'hr',
            'img',
            'input',
            'link',
            'map',
            'meta',
            'noframes',
            'noscript',
            'optgroup',
            'option',
            'param',
            'script',
            'select',
            'style',
            'textarea',
            'title',
        ],
        /**
         * @param {Element} node
         * @param {string[]|null} list
         * @return {boolean}
         */
        test: function (node, list) {
            var elements = list || this.elements;
            return elements.indexOf(node.tagName.toLowerCase()) > -1;
        },
    };
    return NodeUtils;
}());

var RangeUtils = /** @class */ (function () {
    function RangeUtils() {
    }
    RangeUtils.trim = function (range) {
        var oldStart = range.startContainer;
        var newStart = NodeUtils.nextValidTextNode(oldStart, true);
        var newStartOffset = StringUtils.indexOfNthNotEmptyChar(newStart.data, 0);
        if (oldStart === newStart) {
            newStartOffset = Math.max(range.startOffset, newStartOffset);
            if (newStartOffset > StringUtils.indexOfNthNotEmptyCharFromEnd(newStart.data, 0)) {
                // start at the tail of a valid text node
                newStart = NodeUtils.nextValidTextNode(newStart, false);
                newStartOffset = StringUtils.indexOfNthNotEmptyChar(newStart.data, 0);
            }
            else {
                // start at midden of a valid text node
                var head = newStart.substringData(newStartOffset, newStart.length - newStartOffset);
                // num of whitespaces at head
                var headSpaces = StringUtils.indexOfNthNotEmptyChar(head, 0);
                newStartOffset += headSpaces; // skip whitespaces
            }
        }
        var oldEnd = range.endContainer;
        var newEnd = NodeUtils.prevValidTextNode(oldEnd, true);
        var newEndOffset = StringUtils.indexOfNthNotEmptyCharFromEnd(newEnd.data, 0) + 1;
        if (oldEnd === newEnd) {
            newEndOffset = Math.min(range.endOffset, newEndOffset);
            if (newEndOffset <= StringUtils.indexOfNthNotEmptyChar(newEnd.data, 0)) {
                // end at the head of a valid text node
                newEnd = NodeUtils.prevValidTextNode(newEnd, false);
                newEndOffset = StringUtils.indexOfNthNotEmptyCharFromEnd(newEnd.data, 0) + 1;
            }
            else {
                // end at midden of a valid text node
                var tail = newEnd.substringData(0, newEndOffset);
                // num of whitespaces at tail
                var tailSpaces = tail.length - 1 - StringUtils.indexOfNthNotEmptyCharFromEnd(tail, 0);
                newEndOffset -= tailSpaces; // skip whitespaces
            }
        }
        range.setStart(newStart, newStartOffset); // start with a non-whitespace char
        range.setEnd(newEnd, newEndOffset); // end with a non-whitespace char
        return range;
    };
    RangeUtils.extractTextNodes = function (range) {
        var sContainer = range.startContainer;
        var eContainer = range.endContainer;
        var sOffset = range.startOffset;
        var eOffset = range.endOffset;
        if (eContainer.data.length !== eOffset) {
            eContainer.splitText(eOffset);
        }
        if (sOffset !== 0) {
            var newStartContainer = sContainer.splitText(sOffset);
            if (eContainer === sContainer) {
                eContainer = newStartContainer;
                eOffset = eOffset - sOffset;
            }
            sContainer = newStartContainer;
            sOffset = 0;
        }
        range.setStart(sContainer, sOffset);
        range.setEnd(eContainer, eOffset);
        return NodeUtils.getTextNodesBetween(sContainer, eContainer);
    };
    return RangeUtils;
}());

var XSelection$$1 = /** @class */ (function () {
    function XSelection$$1(range, xDocument) {
        this.range = range;
        this.xdoc = xDocument;
    }
    /**
     * create {@link XSelection} from native {@link Selection}
     * @param selection the native `Selection` from which to create the `XSelection` instance.
     * @param xDoc the context of the to be created `XSelection` instance.
     */
    XSelection$$1.fromSelection = function (selection, xDoc) {
        var range = XSelection$$1.rangeFromSelection(selection, xDoc);
        return XSelection$$1.fromRange(range, false, xDoc);
    };
    /**
     * create {@link XSelection} from {@link ITextRange} pojo;
     * {@link ITextRange} specifies the start point and end point
     * <pre>
     * export interface ITextRange {
     *   from: { text: string, nth: number }, // means starting from the `nth` `text`
     *   to: { text: string, nth: number }, // means ending at the `nth` `text`
     * }
     * </pre>
     * @param textRange the serialized pojo from which to create the `XSelection` instance.
     * @param optSelect true to create explicit selection(be selected visibly from browser), false otherwise
     * @param xDoc the context of the to be created `XSelection` instance.
     */
    XSelection$$1.fromTextRange = function (textRange, optSelect, xDoc) {
        if (optSelect === void 0) { optSelect = false; }
        xDoc.refresh();
        var range = XSelection$$1.rangeFromTextRange(textRange, xDoc);
        return XSelection$$1.fromRange(range, optSelect, xDoc);
    };
    /**
     * create {@link XSelection} from {@link ITextRange} pojo
     * @param index the (nth) text to be selected.
     * @param optSelect true to create explicit selection(be selected visibly from browser), false otherwise
     * @param xDoc the context of the to be created `XSelection` instance.
     */
    XSelection$$1.fromText = function (index, optSelect, xDoc) {
        if (optSelect === void 0) { optSelect = false; }
        xDoc.refresh();
        var range = XSelection$$1.rangeFromText(index, xDoc);
        return XSelection$$1.fromRange(range, optSelect, xDoc);
    };
    /**
     * create {@link XSelection} from native {@link Range}
     * @param range the text content to be selected.
     * @param optSelect true to create selection explicitly(be selected visibly from browser), false otherwise
     * @param xDoc the context of the to be created `XSelection` instance.
     */
    XSelection$$1.fromRange = function (range, optSelect, xDoc) {
        if (optSelect === void 0) { optSelect = false; }
        if (range) {
            range = RangeUtils.trim(range);
            if (optSelect) {
                var selection = xDoc.win.getSelection();
                if (selection) {
                    selection.removeAllRanges();
                    selection.addRange(range);
                }
            }
            return new XSelection$$1(range, xDoc);
        }
        return null;
    };
    XSelection$$1.rangeFromSelection = function (selection, xDoc) {
        if (selection.rangeCount > 0 && !selection.isCollapsed) {
            return selection.getRangeAt(0);
        }
        return null;
    };
    XSelection$$1.rangeFromTextRange = function (textRange, xDoc) {
        var head = XSelection$$1.rangeFromText(textRange.start, xDoc);
        var tail = XSelection$$1.rangeFromText(textRange.end, xDoc);
        if (head && tail) {
            var range = xDoc.doc.createRange();
            range.setStart(head.endContainer, head.endOffset);
            range.setEnd(tail.endContainer, tail.endOffset);
            return range;
        }
        return null;
    };
    /**
     *  return an empty range object that has both of its boundary points positioned at
     *  the beginning of the document when no parameters are given, else return an range
     *  object that contains the nth occurrence of the specified text
     *
     * @param {ITextIndex} index
     * @param {XDocument} xDoc
     * @return {Range} the range created that contains the specified text
     * @private
     */
    XSelection$$1.rangeFromText = function (index, xDoc) {
        var text = index.text, nth = index.nth;
        var nodes = xDoc.nodes;
        var cText = StringUtils.compact(text);
        var i = StringUtils.indexOf(xDoc.text, cText, nth || 1);
        if (i > -1) {
            var sPos_1 = i + nodes[0].startPosition; // include the first character
            var ePos_1 = sPos_1 + cText.length; // exclude the last character
            var sIndex = nodes.findIndex(function (node) { return node.endPosition >= sPos_1; });
            var eIndex = nodes.findIndex(function (node) { return node.startPosition >= ePos_1; }) - 1;
            var sContainer = nodes[sIndex];
            var eContainer = eIndex >= 0 ? nodes[eIndex] : nodes[nodes.length - 1];
            if (sContainer) {
                var sOffset = StringUtils.indexOfNthNotEmptyChar(sContainer.data, sPos_1 - sContainer.startPosition);
                var eOffset = StringUtils.indexOfNthNotEmptyChar(eContainer.data, ePos_1 - eContainer.startPosition - 1);
                var range = xDoc.doc.createRange();
                range.setStart(sContainer, sOffset);
                range.setEnd(eContainer, eOffset + 1);
                return range;
            }
        }
        return null;
    };
    XSelection$$1.prototype.trim = function () {
        var selection = this.xdoc.win.getSelection();
        if (selection) {
            selection.removeAllRanges();
            selection.addRange(RangeUtils.trim(this.range));
        }
    };
    /**
     * the text nodes contained in the selection.
     * <em>
     * text node will be splatted into 2 parts if the selection is start from/end in a internal
     * position of a text node
     * </em>
     */
    XSelection$$1.prototype.getTextNodes = function () {
        if (this.texts) {
            return this.texts;
        }
        this.texts = RangeUtils.extractTextNodes(this.range);
        var selection = this.xdoc.win.getSelection();
        if (selection && selection.rangeCount > 0) {
            selection.collapseToStart();
            selection.empty && selection.empty();
            selection.removeAllRanges && selection.removeAllRanges();
        }
        return this.texts;
    };
    /**
     * get the exact "occurrence" of this selection.
     * <em>
     *   the selected text content may appear >=1 times in this page, and the selection is it's
     *   {@code occurrence.nth} time. and the start char of this selection is the
     *   {@code occurrence.position}th char of the whole text content in this page.
     * </em>
     */
    XSelection$$1.prototype.serializeToIndex = function () {
        var cText = StringUtils.compact(this.range.toString());
        if (cText.length < 1) {
            return ITextIndex.None;
        }
        var content = NodeUtils.getContentTo(this.range.endContainer, this.range.endOffset, this.xdoc);
        var texts = StringUtils.find(content, cText);
        if (texts && texts.length > 0) {
            return texts[texts.length - 1];
        }
        return ITextIndex.None;
    };
    /**
     * get the text range of the selection in this page.
     * <pre>
     * export interface ITextRange {
     *   from: { text: string, nth: number }, // means starting from the `nth` `text`
     *   to: { text: string, nth: number }, // means ending at the `nth` `text`
     * }
     * </pre>
     * that is, the text content of this selection starts with {@code textRange.from.text} and
     * ends with {@code textRange.to.text}. and {@code textRange.from.text} is its {@code textRange.from.nth}
     * occurrence, {@code textRange.to.text} is its {@code textRange.to.nth} occurrence.
     * e.g.
     * for text
     * <span>hello world, hello miller, hello world, hello miller, hello miller</span>
     * <pre>
     * {
     *   from:{text: 'hello', nth: 2}
     *   to:{text: 'miller', nth: 3}
     * }
     * means, the selection starts from the 2nd `hello` and ends with 3rd `miller`. that is:
     * <span>hello world, <selected>hello miller, hello world, hello miller, hello miller<selected>, hello world<span>
     * </pre>
     *
     * @param len
     */
    XSelection$$1.prototype.serializeToRange = function (len) {
        if (len === void 0) { len = 5; }
        var cText = StringUtils.compact(this.range.toString());
        if (cText.length < 1) {
            return ITextRange.None;
        }
        var eContent = NodeUtils.getContentTo(this.range.endContainer, this.range.endOffset, this.xdoc);
        var eText = eContent.length > len ? eContent.substr(eContent.length - len, len) : eContent;
        var eTexts = StringUtils.find(eContent, eText);
        var sContent = NodeUtils.getContentTo(this.range.startContainer, this.range.startOffset, this.xdoc);
        var sText = sContent.length > len ? sContent.substr(sContent.length - len, len) : sContent;
        var sTexts = StringUtils.find(sContent, sText);
        if (sTexts && sTexts.length > 0 && eTexts && eTexts.length > 0) {
            return {
                start: sTexts[sTexts.length - 1],
                end: eTexts[eTexts.length - 1]
            };
        }
        return ITextRange.None;
    };
    /**
     * get the selected text content
     * @param compact true to remove the white spaces. false otherwise
     */
    XSelection$$1.prototype.getContent = function (compact) {
        if (compact === void 0) { compact = false; }
        if (compact) {
            return StringUtils.compact(this.range.toString());
        }
        return this.range.toString();
    };
    /**
     * get the native selection
     */
    XSelection$$1.prototype.toSelection = function () {
        var selection = this.xdoc.win.getSelection();
        if (selection.rangeCount < 1) {
            selection.addRange(this.range);
        }
        return selection;
    };
    /**
     * clear the selection
     */
    XSelection$$1.prototype.empty = function () {
        if (!this.texts || this.texts.length < 1) {
            return;
        }
        var first = this.range.startContainer;
        var last = this.range.endContainer;
        if (first && first.parentNode) {
            first.parentNode.normalize();
        }
        if (last && last.parentNode) {
            last.parentNode.normalize();
        }
        this.range.collapse();
        this.texts = [];
        var selection = this.xdoc.win.getSelection();
        selection.collapseToStart();
        selection.empty && selection.empty();
        selection.removeAllRanges && selection.removeAllRanges();
    };
    /**
     * check if the selection is cleared or not.
     */
    XSelection$$1.prototype.isEmpty = function () {
        var selection = this.xdoc.win.getSelection();
        return !selection || selection.rangeCount < 1 || selection.isCollapsed || this.range.collapsed;
    };
    return XSelection$$1;
}());

var XDocument = /** @class */ (function () {
    function XDocument(root) {
        this.selection = null;
        this.root = root;
        this.doc = (root instanceof Document ? root : root.ownerDocument);
        this.win = this.doc.defaultView;
        this.refresh();
    }
    XDocument.from = function (root) {
        if (root === void 0) { root = document.body; }
        return new XDocument(root);
    };
    XDocument.prototype.refresh = function () {
        var nodes = NodeUtils.getValidTextNodes(this.root);
        var position = 0;
        var text = '';
        nodes.forEach(function (node) {
            node.startPosition = position;
            var nodeText = StringUtils.compact(node.data);
            text += nodeText;
            position += nodeText.length;
            node.endPosition = position - 1;
        });
        this.text = text;
        this.nodes = nodes;
    };
    /**
     * @see {@code XSelection.fromText}
     */
    XDocument.prototype.select = function (optText, optNth, optSelect) {
        if (optNth === void 0) { optNth = 1; }
        if (optSelect === void 0) { optSelect = false; }
        if (optText && optText.trim()) {
            if (this.selection && !this.selection.isEmpty()) {
                this.selection.empty();
            }
            this.selection = this.fromText(optText.trim(), optNth, optSelect);
        }
        else {
            var selection = this.win.getSelection();
            if (selection) {
                this.selection = this.fromSelection(selection);
            }
        }
        return this.selection;
    };
    /**
     * @see {@code XSelection.fromSelection}
     */
    XDocument.prototype.fromSelection = function (selection) {
        return this.selection = XSelection$$1.fromSelection(selection, this);
    };
    /**
     * @see {@code XSelection.fromText}
     */
    XDocument.prototype.fromText = function (text, nth, optSelect) {
        if (nth === void 0) { nth = 1; }
        if (optSelect === void 0) { optSelect = false; }
        return this.selection = this.fromTextIndex({ text: text, nth: nth }, optSelect);
    };
    /**
     * @see {@code XSelection.fromText}
     */
    XDocument.prototype.fromTextIndex = function (index, optSelect) {
        if (optSelect === void 0) { optSelect = false; }
        return this.selection = XSelection$$1.fromText(index, optSelect, this);
    };
    /**
     * @see {@code XSelection.fromTextRange}
     */
    XDocument.prototype.fromTextRange = function (range, optSelect) {
        if (optSelect === void 0) { optSelect = false; }
        return this.selection = XSelection$$1.fromTextRange(range, optSelect, this);
    };
    XDocument.prototype.normalize = function () {
        var _this = this;
        this.root.normalize();
        var end = 0;
        this.nodes.reverse().forEach(function (node, index) {
            if (!node.parentNode) {
                end = node.endPosition;
                _this.nodes.splice(_this.nodes.length - index, 1);
            }
            else if (end > 0) {
                node.endPosition = end;
            }
        });
    };
    return XDocument;
}());

var ITextIndex;
(function (ITextIndex) {
    ITextIndex.None = {
        nth: -1,
        position: -1,
        text: '',
    };
})(ITextIndex || (ITextIndex = {}));
var ITextRange;
(function (ITextRange) {
    ITextRange.None = {
        start: { text: '', nth: -1 },
        end: { text: '', nth: -1 },
    };
    function toSimpleString(range) {
        return JSON.stringify([range.start.nth + ":" + range.start.text, range.end.nth + ":" + range.end.text]);
    }
    ITextRange.toSimpleString = toSimpleString;
    function fromSimpleString(str) {
        var simple = JSON.parse(str);
        var from = simple[0].split(':', 2);
        var to = simple[1].split(':');
        return {
            start: { text: from[1], nth: parseInt(from[0], 10) },
            end: { text: to[1], nth: parseInt(to[0], 10) }
        };
    }
    ITextRange.fromSimpleString = fromSimpleString;
})(ITextRange || (ITextRange = {}));

export { ITextIndex, ITextRange, XSelection$$1 as XSelection, XDocument, NodeUtils };
//# sourceMappingURL=texton.es5.js.map
